import numpy as np
import skimage.measure as smeasure


def pixel2mask(image: np.ndarray, low: float, high: float) -> np.ndarray:
    """
    Binarize pixel intensities. Pixels with intensity higher than the lower
    bound and also connected to at least one pixel of intensity higher than
    higher bound will be labeled True, otherwise False.

    Args:
        image: np.ndarray
            Intensity image.
        low: float
            Lower bound.
        high: float
            Higher bound.

    Return: np.ndarray of type bool
    """
    mask = image > low
    labels = smeasure.label(mask, background=0)
    for region in smeasure.regionprops(label_image=labels, intensity_image=image):
        if region.max_intensity < high:
            mask[region.coords[:, 0], region.coords[:, 1]] = 0

    return mask



from pathlib import Path


def check_overwrite(overwrite: bool, path: Path):
    """
    Check file existence and either raise exception or delete it.

    Args:
        overwrite: bool
            Overwrite flag.
        path: pathlib.Path object
            File/folder path to check.
    """
    if overwrite:
        if path.is_file():
            path.unlink()
        elif path.is_dir():
            path.rmdir()
    elif path.exists():
        raise ValueError(f"{path} exists but overwrite is set to {overwrite}")



# `blockwise_view.py` from `lazyflow` module of `ilastik` project
# https://github.com/ilastik/ilastik/blob/master/lazyflow/utility/blockwise_view.py

from __future__ import division
from builtins import map
import numpy

try:
    # If you use vigra, we do special handling to preserve axistags
    import vigra

    _vigra_available = True
except ImportError:
    _vigra_available = False


def blockwise_view(a, blockshape, aslist=False, require_aligned_blocks=True):
    """
    Return a 2N-D view of the given N-D array, rearranged so each ND block (tile)
    of the original array is indexed by its block address using the first N
    indexes of the output array.

    Note: This function is nearly identical to ``skimage.util.view_as_blocks()``, except:
          - "imperfect" block shapes are permitted (via require_aligned_blocks=False)
          - only contiguous arrays are accepted.  (This function will NOT silently copy your array.)
            As a result, the return value is *always* a view of the input.

    Args:
        a: The ND array

        blockshape: The tile shape

        aslist: If True, return all blocks as a list of ND blocks
                instead of a 2D array indexed by ND block coordinate.

        require_aligned_blocks: If True, check to make sure no data is "left over"
                                in each row/column/etc. of the output view.
                                That is, the blockshape must divide evenly into the full array shape.
                                If False, "leftover" items that cannot be made into complete blocks
                                will be discarded from the output view.

    Here's a 2D example (this function also works for ND):

    >>> a = numpy.arange(1,21).reshape(4,5)
    >>> print(a)
    [[ 1  2  3  4  5]
     [ 6  7  8  9 10]
     [11 12 13 14 15]
     [16 17 18 19 20]]

    >>> view = blockwise_view(a, (2,2), require_aligned_blocks=False)
    >>> print(view)
    [[[[ 1  2]
       [ 6  7]]
    <BLANKLINE>
      [[ 3  4]
       [ 8  9]]]
    <BLANKLINE>
    <BLANKLINE>
     [[[11 12]
       [16 17]]
    <BLANKLINE>
      [[13 14]
       [18 19]]]]

    Inspired by the 2D example shown here: http://stackoverflow.com/a/8070716/162094
    """
    assert a.flags["C_CONTIGUOUS"], "This function relies on the memory layout of the array."
    blockshape = tuple(blockshape)
    outershape = tuple(numpy.array(a.shape) // blockshape)
    view_shape = outershape + blockshape

    if require_aligned_blocks:
        assert (
            numpy.mod(a.shape, blockshape) == 0
        ).all(), "blockshape {} must divide evenly into array shape {}".format(blockshape, a.shape)

    # inner strides: strides within each block (same as original array)
    intra_block_strides = a.strides

    # outer strides: strides from one block to another
    inter_block_strides = tuple(a.strides * numpy.array(blockshape))

    # This is where the magic happens.
    # Generate a view with our new strides (outer+inner).
    view = numpy.lib.stride_tricks.as_strided(a, shape=view_shape, strides=(inter_block_strides + intra_block_strides))

    # Special handling for VigraArrays
    if _vigra_available and isinstance(a, vigra.VigraArray) and hasattr(a, "axistags"):
        view_axistags = vigra.AxisTags([vigra.AxisInfo() for _ in blockshape] + list(a.axistags))
        view = vigra.taggedView(view, view_axistags)

    if aslist:
        return list(map(view.__getitem__, numpy.ndindex(outershape)))
    return view


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# modified from https://gist.github.com/jmuhlich/a926f55f7eb115af54c9d4754539bbc1

from __future__ import print_function, division
import sys
import os
import re
import io
import struct
import itertools
import uuid
import multiprocessing
import concurrent.futures
import numpy as np
import skimage.io
import tifffile
import skimage.transform
import pytiff


def preduce(coords, img_in, img_out):
    (iy1, ix1), (iy2, ix2) = coords
    (oy1, ox1), (oy2, ox2) = np.array(coords) // 2
    tile = img_in[iy1:iy2, ix1:ix2]
    #    tile = skimage.transform.pyramid_reduce(tile, multichannel=False)
    tile = skimage.transform.downscale_local_mean(tile, (2, 2)).astype(img_out.dtype)
    #    with warnings.catch_warnings():
    #        warnings.filterwarnings('ignore', 'Possible precision loss')
    #        tile = skimage.util.dtype.convert(tile, img_out.dtype)
    img_out[oy1:oy2, ox1:ox2] = tile


def imsave(path, img, tile_size, **kwargs):
    tifffile.imsave(
        path,
        img,
        bigtiff=True,
        append=True,
        tile=(tile_size, tile_size),
        metadata=None,
        **kwargs
    )


def format_shape(shape):
    return "%dx%d" % (shape[1], shape[0])


def construct_xml(
    filename, shapes, num_channels, dtype, pixel_size=1, channel_name_list=None
):
    if channel_name_list is None:
        channel_name_list = ["Channel {}".format(i) for i in range(num_channels)]

    img_uuid = uuid.uuid4().urn
    if dtype == np.uint16:
        ome_dtype = "uint16"
    elif dtype == np.uint8:
        ome_dtype = "uint8"
    else:
        raise ValueError("can't handle dtype: %s" % dtype)
    ifd = 0
    xml = io.StringIO()
    xml.write(u'<?xml version="1.0" encoding="UTF-8"?>')
    xml.write(
        (
            u'<OME xmlns="http://www.openmicroscopy.org/Schemas/OME/2016-06"'
            ' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'
            ' UUID="{uuid}"'
            ' xsi:schemaLocation="http://www.openmicroscopy.org/Schemas/OME/2016-06'
            ' http://www.openmicroscopy.org/Schemas/OME/2016-06/ome.xsd">'
        ).format(uuid=img_uuid)
    )
    for level, shape in enumerate(shapes):
        if level == 0:
            psize_xml = (
                u'PhysicalSizeX="{0}" PhysicalSizeXUnit="\u00b5m"'
                u' PhysicalSizeY="{0}" PhysicalSizeYUnit="\u00b5m"'.format(pixel_size)
            )
        else:
            psize_xml = u""
        xml.write(u'<Image ID="Image:{}">'.format(level))
        xml.write(
            (
                u'<Pixels BigEndian="false" DimensionOrder="XYZCT"'
                ' ID="Pixels:{level}" {psize_xml} SizeC="{num_channels}" SizeT="1"'
                ' SizeX="{sizex}" SizeY="{sizey}" SizeZ="1" Type="{ome_dtype}">'
            ).format(
                level=level,
                psize_xml=psize_xml,
                num_channels=num_channels,
                sizex=shape[1],
                sizey=shape[0],
                ome_dtype=ome_dtype,
            )
        )
        for channel, channel_name in zip(range(num_channels), channel_name_list):
            xml.write(
                (
                    u'<Channel ID="Channel:{level}:{channel}"'
                    + (u' Name="{channel_name}"' if level == 0 else u"")
                    + u' SamplesPerPixel="1"><LightPath/></Channel>'
                ).format(level=level, channel=channel, channel_name=channel_name)
            )
        for channel in range(num_channels):
            xml.write(
                (
                    u'<TiffData FirstC="{channel}" FirstT="0" FirstZ="0"'
                    ' IFD="{ifd}" PlaneCount="1">'
                    '<UUID FileName="{filename}">{uuid}</UUID>'
                    "</TiffData>"
                ).format(channel=channel, ifd=ifd, filename=filename, uuid=img_uuid)
            )
            ifd += 1
        if level == 0:
            for channel in range(num_channels):
                xml.write(
                    u'<Plane TheC="{channel}" TheT="0" TheZ="0"/>'.format(
                        channel=channel
                    )
                )
        xml.write(u"</Pixels>")
        xml.write(u"</Image>")
    xml.write(u"</OME>")
    xml_bytes = xml.getvalue().encode("utf-8") + b"\x00"
    return xml_bytes


def patch_ometiff_xml(path, xml_bytes):
    with open(path, "rb+") as f:
        f.seek(0, io.SEEK_END)
        xml_offset = f.tell()
        f.write(xml_bytes)
        f.seek(0)
        ifd_block = f.read(500)
        match = re.search(b"!!xml!!\x00", ifd_block)
        if match is None:
            raise RuntimeError("Did not find placeholder string in IFD")
        f.seek(match.start() - 8)
        f.write(struct.pack("<Q", len(xml_bytes)))
        f.write(struct.pack("<Q", xml_offset))


def build_pyramid(array_list, channel_name_list=None, out_path=None, tile_size=1024):
    if hasattr(os, "sched_getaffinity"):
        num_workers = len(os.sched_getaffinity(0))
    else:
        num_workers = multiprocessing.cpu_count()

    if channel_name_list is None:
        channel_name_list = [str(i + 1) for i in range(len(array_list))]

    if out_path is None:
        out_path = "./out.ome.tif"

    if os.path.exists(out_path):
        print("%s already exists, aborting" % out_path)
        sys.exit(1)

    print("Appending input images")
    for i, img_in in enumerate(array_list):
        print("    %d: %s" % (i + 1, channel_name_list[i]))
        if i == 0:
            base_shape = img_in.shape
            dtype = img_in.dtype
            kwargs = {"description": "!!xml!!", "software": "Glencoe/Faas pyramid"}
        else:
            if img_in.shape != base_shape:
                print(
                    "%s: expected shape %s, got %s"
                    % (channel_name_list[i], base_shape, img_in.shape)
                )
                sys.exit(1)
            if img_in.dtype != dtype:
                print(
                    "%s: expected dtype %s, got %s"
                    % (channel_name_list[i], dtype, img_in.dtype)
                )
                sys.exit(1)
            kwargs = {}
        imsave(out_path, img_in, tile_size, **kwargs)
    print()

    num_channels = len(channel_name_list)
    num_levels = np.ceil(np.log2(max(base_shape) / tile_size)) + 1
    factors = 2 ** np.arange(num_levels)
    shapes = (np.ceil(np.array(base_shape) / factors[:, None])).astype(int)

    print("Pyramid level sizes:")
    for i, shape in enumerate(shapes):
        print("    level %d: %s" % (i + 1, format_shape(shape)), end="")
        if i == 0:
            print(" (original size)", end="")
        print()
    print()

    executor = concurrent.futures.ThreadPoolExecutor(num_workers)

    shape_pairs = zip(shapes[:-1], shapes[1:])
    for level, (shape_in, shape_out) in enumerate(shape_pairs):

        print(
            "Resizing images for level %d (%s -> %s)"
            % (level + 2, format_shape(shape_in), format_shape(shape_out))
        )

        ty = np.array(range(0, shape_in[0], tile_size))
        tx = np.array(range(0, shape_in[1], tile_size))
        coords = list(
            zip(
                itertools.product(ty, tx),
                itertools.product(ty + tile_size, tx + tile_size),
            )
        )
        img_in = pytiff.Tiff(out_path)
        img_out = np.empty(shape_out, dtype)

        for c in range(num_channels):

            page = level * num_channels + c
            img_in.set_page(page)
            for i, _ in enumerate(
                executor.map(
                    preduce, coords, itertools.repeat(img_in), itertools.repeat(img_out)
                )
            ):
                percent = int((i + 1) / len(coords) * 100)
                if i % 20 == 0 or percent == 100:
                    print("\r    %d: %d%%" % (c + 1, percent), end="")
                    sys.stdout.flush()
            imsave(out_path, img_out, tile_size)
            print()

        img_in.close()
        print()

    xml = construct_xml(
        os.path.basename(out_path),
        shapes,
        num_channels,
        dtype,
        0.325,
        channel_name_list=channel_name_list,
    )
    patch_ometiff_xml(out_path, xml)



import typing

import numpy as np
import pandas as pd

from skimage.measure import regionprops

from external import blockwise_view


def minitile_corrcoef(
    arr1: np.ndarray,
    arr2: np.ndarray,
    block_shape: typing.Tuple[int, int],
    keep_shape: bool = False,
) -> np.ndarray:
    """
    Blockwise correlation coefficient.

    Args:
        arr1, arr2: np.ndarray
            Input images to compare.
        block_shape: tuple of int
            Shape of mini-tile.
        keep_shape: bool [optional]
            Keep same shape as input images or just one pixel per mini-tile.
            Default False.

    Return: correlation coefficient heatmap.
    """
    # make blocks
    b1 = blockwise_view(arr1, block_shape, require_aligned_blocks=False, aslist=False)
    b2 = blockwise_view(arr2, block_shape, require_aligned_blocks=False, aslist=False)
    # subtract mean
    axes = tuple(np.arange(b1.ndim, dtype=int)[b1.ndim // 2:])
    b1 -= b1.mean(axis=axes, keepdims=True)
    b2 -= b2.mean(axis=axes, keepdims=True)
    # numerator of corrcoef
    numerator = np.multiply(b1, b2).mean(axis=axes, keepdims=keep_shape)
    # denomenator of corrcoef
    dof = np.prod(b1.shape[slice(axes[0], axes[-1] + 1)])
    b1_std = np.sqrt((b1 ** 2).mean(axis=axes, keepdims=keep_shape) / dof)
    b2_std = np.sqrt((b2 ** 2).mean(axis=axes, keepdims=keep_shape) / dof)
    denominator = np.multiply(b1_std, b2_std)
    # divide
    out = np.divide(numerator, denominator)

    return out


def region_corrcoef(
    arr1: np.ndarray, arr2: np.ndarray, mask: np.ndarray, return_dataframe: bool = True
) -> typing.Union[np.ndarray, pd.DataFrame]:
    """
    Region-wise correlation coefficient.

    Args:
        arr1, arr2: np.ndarray
            Input images to compare.
        mask: np.ndarray
            Mask of integer defining regions.
        return_dataframe: bool [optional]
            If True, return dataframe of region label and correlation
            coefficients, else return an array with same shape as input.
            Default True.
    """
    # if memory usage is an issue here, pass dask.array as input
    # and use dask_image.ndmeasure.labeled_comprehension
    coef_dict = {}
    for region in regionprops(mask):
        pix1 = arr1[region.coords[:, 0], region.coords[:, 1]]
        pix2 = arr2[region.coords[:, 0], region.coords[:, 1]]
        coef_dict[region.label] = np.corrcoef(pix1, pix2)[0, 1]

    if return_dataframe:
        df = pd.DataFrame.from_records(
            list(coef_dict.items()), columns=["label", "corrcoef"]
        )
        return df
    else:
        coef_dict[0] = 0.0
        coef_image = np.vectorize(lambda x: coef_dict[x])(mask)
        return coef_image



import typing
import collections
from pathlib import Path

import h5py
import tifffile

from .util import check_overwrite
from .external import ashlar_pyramid


def uniquify(name_list: typing.List[str]) -> typing.List[str]:
    """
    Uniquify names, a common issue in markers.csv files seen so far.

    Args:
        name_list: list of str
            Input names.

    Return: list of str, each being unique
    """
    count = collections.Counter(name_list)
    if count.most_common(1)[1] == 1:
        # most common name has frequency of one, meaning all names unique
        return name_list
    else:
        dup_count = {c[0]: 0 for c in count.most_common() if c[1] > 1}
        for i, name in enumerate(name_list):
            if name in dup_count:
                dup_count[name] += 1
                name_list[i] = f"{name}_{dup_count[name]}"
        return name_list


def name_generator(start: int = 0, step: int = 1):
    """
    Ever-increasing number generator.

    Args:
        start: int [optional]
            Starting number, default 0.
        step: int [optional]
            Increment, default 1.
    """
    x = start
    while True:
        yield str(x)
        x += step


def ometif2tif(
    in_filepath: str,
    out_folderpath: str,
    names_filepath: str = None,
    overwrite: bool = True,
):
    """
    Unpack ome.tif file into a folder of TIFF images.

    Args:
        in_filepath: str
            Input ome.tif file path.
        out_folderpath: str
            Output folder path.
        name_filepath: str [optional]
            Text file with each line the name of the channels.
        overwrite: bool [optional]
            Overwrite output folder if already exists. Default True.
    """
    # preprocessing
    in_filepath = Path(in_filepath)
    out_folderpath = Path(out_folderpath)
    names_filepath = Path(names_filepath)
    check_overwrite(overwrite=overwrite, path=out_folderpath)

    # load marker name
    if names_filepath is None:
        name_list = name_generator(start=1, step=1)
    else:
        name_list = names_filepath.read_text().splitlines()
        name_list = uniquify(name_list)

    # try to minimize memory usage here as each image can be several GB
    with tifffile.TiffFile(in_filepath) as tif:
        for name, pg in zip(name_list, tif.series[0].pages):
            out_filepath = out_folderpath / f"{name}.tif"
            tifffile.imsave(out_filepath, pg.asarray())


def ometif2hdf5(
    in_filepath: str,
    out_filepath: str,
    names_filepath: str = None,
    overwrite: bool = True,
):
    """
    Convert ome.tif file into HDF5 file.

    Args:
        in_filepath: str
            Input ome.tif file path.
        out_filepath: str
            Output .h5 file path.
        name_filepath: str [optional]
            Text file with each line the name of the channels. Default None.
            If None, as increasing numbers like 1, 2, 3, ...
        overwrite: bool [optional]
            Overwrite output folder if already exists. Default True.
    """
    # preprocessing
    in_filepath = Path(in_filepath)
    out_filepath = Path(out_filepath)
    names_filepath = Path(names_filepath)
    check_overwrite(overwrite=overwrite, path=out_filepath)

    # load marker name
    if names_filepath is None:
        name_list = name_generator(start=1, step=1)
    else:
        name_list = names_filepath.read_text().splitlines()
        name_list = uniquify(name_list)

    # automatic chunk size estimation
    with tifffile.TiffFile(in_filepath) as tif, h5py.File(out_filepath, "w") as out_f:
        for name, pg in zip(name_list, tif.series[0].pages):
            out_f.create_dataset(name=name, data=pg.asarray())


def tif2ometif(
    in_folderpath: str, out_filepath: str, tile_size: int = 1024, overwrite: bool = True
):
    """
    Concatenate channels and make pyramid.

    Args:
        in_folderpath: str
            Folder of input tif images.
        out_filepath: str
            Output ome.tif file path.
        tile_size: int [optional]
            Tile size lower cap for ashlar pyramid generator. Default 1024.
        overwrite: bool [optional]
            Overwrite flag, default True.
    """
    # preprocessing
    in_folderpath = Path(in_folderpath)
    out_filepath = Path(out_filepath)
    check_overwrite(overwrite=overwrite, path=out_filepath)

    # make generators
    def arr_gen():
        for x in in_folderpath.iterdir():
            yield tifffile.imread(x)

    def name_gen():
        for x in in_folderpath.iterdir():
            yield x.stem

    # pass to ashlar pyramid generator
    ashlar_pyramid.main(
        array_list=arr_gen(),
        channel_name_list=name_gen(),
        out_path=out_filepath,
        tile_size=tile_size,
    )



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pcatk",
    version="0.1",
    author="Hung-Yi Wu",
    author_email="hungyi_wu@g.harvard.edu",
    description="Pre-Cancer Atlas data analysis toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hungyiwu/pca_analysis_toolkit",
    packages=setuptools.find_packages(),
    install_requires=[
        "numpy",
        "scikit-image",
        ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        ],
    )



import typing
from pathlib import Path

import numpy as np
import skimage.io as sio
import matplotlib.colors as mcolors

from skimage.measure import regionprops_table
from skimage import img_as_float, img_as_ubyte

from .util import check_overwrite


Slice = typing.Tuple[slice, slice]


def sample(
    mask: np.ndarray, tile_size: typing.Tuple[int, int], size: int
) -> typing.Dict[int, Slice]:
    """
    Sample cells.

    Args:
        mask: np.ndarray
            Mask image for skimage.feature.regionprops
        tile_size: tuple of int
            Crop size for individual cells.
        size: int
            Sample size.

    Return: dict of label -> tuple of slice objects
    """
    # filter out cells at boundary
    props = regionprops_table(mask, properties=["label", "centroid"])
    mask = (
        (props["centroid-0"] > tile_size[0])
        & (props["centroid-0"] < mask.shape[0] - tile_size[0])
        & (props["centroid-1"] > tile_size[1])
        & (props["centroid-1"] < mask.shape[1] - tile_size[1])
    )
    for key in props:
        props[key] = props[key][mask]
    # sampling
    count = props["label"].shape[0]
    index = np.random.choice(range(count), size=size, replace=False)
    for key in props:
        props[key] = props[key][index]
    # make slice output
    xl = np.rint(props["centroid-0"] - tile_size[0] // 2).astype(int)
    yl = np.rint(props["centroid-1"] - tile_size[1] // 2).astype(int)
    xu, yu = xl + tile_size[0], yl + tile_size[1]
    out_dict = {
        props["label"][i]: (slice(xl[i], xu[i]), slice(yl[i], yu[i]))
        for i in range(size)
    }

    return out_dict


def project2rgb(image: np.ndarray, color: str) -> np.ndarray:
    """
    Project an image from matplotlib.colors named color to RGB space.

    Args:
        image: np.ndarray
            Input 2-D image of shape (N, M).
        color: str
            matplotlib.colors named color.

    Return: np.ndarray of shape (N, M, 3)
    """
    out = color.gray2rgb(img_as_float(image))
    rgb_code = mcolors.to_rgb(color)

    return out * np.array(rgb_code)


def render(
    slice_dict: typing.Dict[int, Slice],
    image_dict: typing.Dict[str, np.ndarray],
    out_folderpath: str,
    overwrite: bool = True,
):
    """
    Render sampled cells.

    Args:
        slice_dict: dict of integer --> tuple of slice object.
            Output of sample function.
        image_dict: dict of str --> np.ndarray
            Key is matplotlib named color. Value is image as np.ndarray.
        out_folderpath: str
            Folder to save rendered figures.
        overwrite: bool [optional]
            Overwrite flag, default True.
    """
    # preprocessing
    out_folderpath = Path(out_folderpath)
    check_overwrite(overwrite=overwrite, path=out_folderpath)

    # not much memory optimization done for now
    # if memory is an issue, pass h5py.Dataset instead of np.ndarray
    # code below is compatible with h5py.Dataset
    for key in slice_dict:
        arr = np.stack(
            [project2rgb(image, color) for color, image in image_dict.items()], axis=0
        )
        arr = arr.max(axis=0)
        arr = img_as_ubyte(arr)
        out_filepath = out_folderpath / f"{key}.png"
        sio.imsave(out_filepath, arr)


def assemble(
    in_folderpath: str,
    out_filepath: str,
    shape: typing.Tuple[int, int],
    pad_width: int = None,
    overwrite: bool = True,
):
    """
    Assemble images into exemplar, supports padding.

    Args:
        in_folderpath: str
            Folder path of input images.
        out_filepath: str
            Output image path.
        shape: tuple of int
            Mosaic shape, ex. (10, 10).
        pad_width: int [optional]
            Padding as border. If None, no padding. Default None.
        overwrite: bool [optional]
            Overwrite flag, default True.
    """
    # preprocessing
    in_folderpath = Path(in_folderpath)
    out_filepath = Path(out_filepath)
    check_overwrite(out_filepath)

    # if fewer image than needed, add blank images
    image_list = [sio.imread(p) for p in in_folderpath.iterdir()]
    num_block = np.prod(shape)
    if len(image_list) < num_block:
        image_list.extend(
            [np.ones_like(image_list[0]) for _ in range(num_block - len(image_list))]
        )

    # add padding
    if pad_width is not None:
        image_list = [
            np.pad(
                image,
                pad_width=(pad_width, pad_width, 0),
                mode="constant",
                constant_values=1,
            )
            for image in image_list
        ]

    # assembling
    image_list = [image_list[i:i + shape[0]] for i in range(0, num_block, shape[0])]
    out = np.block(image_list)
    sio.imsave(out_filepath, out)
